
#ifdef INTERNEL_ERROR
typedef Vec<16, int> IndexList ;
typedef Vec<16, int> TargetList ;
#else
typedef Vec<int, 16> IndexList ;
typedef Vec<int, 16> TargetList ;
#endif

extern int InverseRemp;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `MessSchedule' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class MessSchedule {
public :
  void gather(void* locBase, void* srvBase) ;
  void scatter(void* locBase, void* srvBase) ;
  void combine(void* locBase, void* srvBase, void (*op)(void*, void*)) ;
protected :
  MessSchedule(const int _elementLen) ;
    // add by Zheng Gengbin
  MessSchedule(const int _elementLen, Status _locStatus, Status _srvStatus) ;
  ~MessSchedule() ;

  void sendReq(int offset, int dstId) { (srvIdxs [dstId]).add(offset) ; }
  void recvReq(int offset, int srcId) { (locIdxs [srcId]).add(offset) ; }

  void build() ;

  int elementLen ;

  int nProcs, locId ;

  IndexList* locIdxs ;
  IndexList* srvIdxs ;

  TargetList locList ;
  TargetList srvList ;

  char* locBuf ;
  char* srvBuf ;

     // Zheng Gengbin
  Status srvStatus, locStatus ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `SendServer' class.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class SendServer : public Server {
public :
  SendServer(const int numMes, int* remIds,
             char* _sendBuf, IndexList* _sendIdxs,
             void* _sendBase, int _elementLen, Status _status=IN_CORE) ;  // add by Zheng
  ~SendServer() ;

  void transition(MPI_Status* status) ;

  int active() const ;

  void initSend(const int remId) ;
private :
  int numMes ;
  int *remIds ;

  char* sendBuf ;
  IndexList* sendIdxs ;

  void* sendBase ;
  int elementLen ;

  int mes ;
 
    // Zheng Gengbin
  Status status;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockMessSchedule' class.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockMessSchedule : protected GlobalEnv {
public :
  BlockMessSchedule(const int _elementLen, const int _rank) :
      locMessages(nProcs), srvMessages(nProcs), // cpyNum(0),
      locMemCopy(4), srvMemCopy(4), 
      elementLen(_elementLen), rank(_rank), srvStatus(IN_CORE), locStatus(IN_CORE) 
      {locStorage=srvStorage=defaultStorageMode;};

    // new Constructor for OUT_OF_CORE
  BlockMessSchedule(const int _elementLen, const int _rank, 
		    const Status locs, const Status srvs, 
		    const int _locStorage, const int _srvStorage) :
      locMessages(nProcs), srvMessages(nProcs), 
      locMemCopy(4), srvMemCopy(4), 
      elementLen(_elementLen), rank(_rank), srvStatus(locs), locStatus(srvs),
      srvStorage(_locStorage), locStorage(_srvStorage) {};

  void sendReq(int offset, int* strs, int ctg,
	       int* exts, int vol, const int dstId) {
    if(dstId == locId) {
	// keep code here
      //addBlock(srvCopies, offset, strs, ctg, elementLen * vol, exts, rank, srvStatus) ;
      //cpyNum++ ;

      const int hop = 4;
      MessQueue* queue = srvMemCopy.getElement(hop) ;
	// add by Zheng Gengbin
//AdlibPrintf("sendReq:local.\n");
      addMessageBlock(queue->messages, queue->blocks, offset, strs, ctg, 
                      elementLen * vol, exts, rank, srvStatus, IclaSize, srvStorage) ;
      queue->remId = dstId ;
    }
    else {
//      const int hop = (dstId - locId + nProcs) % nProcs ;
      int hop ;
      if ( InverseRemp == 0) 
	hop = (dstId - locId + nProcs) % nProcs ;
      else
	hop = (locId - dstId + nProcs) % nProcs ;
      MessQueue* queue = srvMessages.getElement(hop) ;
	// add by Zheng Gengbin
//AdlibPrintf("sendReq:send.\n");
      addMessageBlock(queue->messages, queue->blocks, offset, strs, ctg, 
		      elementLen * vol, exts, rank, srvStatus, 
		      IclaSize<STATIC_BUFFER_SIZE?IclaSize:STATIC_BUFFER_SIZE, srvStorage) ;
      queue->remId = dstId ;
    }
  }

  void recvReq(int offset, int* strs, int ctg,
	       int* exts, int vol, const int srcId) {
    if(srcId == locId) {
	// keep code here
      //addBlock(locCopies, offset, strs, ctg, elementLen * vol, exts, rank, locStatus) ;

      const int hop = 4;
      MessQueue* queue = locMemCopy.getElement(hop) ;
	// add by Zheng Gengbin
//AdlibPrintf("recvReq:local.\n");
      addMessageBlock(queue->messages, queue->blocks, offset, strs, ctg, 
		      elementLen * vol, exts, rank, locStatus, IclaSize, locStorage);
      queue->remId = srcId ;
    }
    else {
//      const int hop = (locId - srcId + nProcs) % nProcs ;
      int hop ;
      if ( InverseRemp == 0) 
	hop = (locId - srcId + nProcs) % nProcs ;
      else
	hop = (srcId - locId + nProcs) % nProcs ;
      MessQueue* queue = locMessages.getElement(hop) ;
	// add by Zheng Gengbin
//AdlibPrintf("recvReq:recv.\n");
      addMessageBlock(queue->messages, queue->blocks, offset, strs, ctg, 
                      elementLen * vol, exts, rank, locStatus, IclaSize<STATIC_BUFFER_SIZE?IclaSize:STATIC_BUFFER_SIZE, locStorage) ;
      queue->remId = srcId ;
    }
  }

  void build() {
    locList = locMessages.toList() ;
    srvList = srvMessages.toList() ;
  }

/*
   old code
  void gather(void* locBase, void* srvBase) {
    run(locList, srvList, &locCopies, &srvCopies, locBase, srvBase) ;
  }
*/

    // add by Zheng Gengbin
  void gather(void* locBase, void* srvBase) {
    run(locList, srvList, &locMemCopy, &srvMemCopy, locBase, srvBase) ;
  }

/*
   old code
  void scatter(void* locBase, void* srvBase) {
    run(srvList, locList, &srvCopies, &locCopies, srvBase, locBase) ;
  }
*/

    // add by Zheng Gengbin
  void scatter(void* locBase, void* srvBase) {
    run(srvList, locList, &srvMemCopy, &locMemCopy, srvBase, locBase) ;
  }

  void show() ;

protected :

//wjp add
  Status srvStatus, locStatus ; 
  int srvStorage, locStorage;

  int elementLen, rank ;
  int srcType, dstType;

  SparseArray<MessQueue> locMessages, srvMessages ;
  List<MessQueue> *locList, *srvList ;

    // these are used by subclass !
  MultiBlockList locCopies, srvCopies ;
  int cpyNum ;

  SparseArray<MessQueue> locMemCopy, srvMemCopy ;

private :
/*
  void run(List<MessQueue>* dstMessages, List<MessQueue>* srcMessages,
           MultiBlockList* dstCopies, MultiBlockList* srcCopies,
           void* dstBase, void* srcBase) ;
*/

    // add by Zheng Gengbin
  void run(List<MessQueue>* dstMessages, List<MessQueue>* srcMessages,
           SparseArray<MessQueue>* dstMemCopy, SparseArray<MessQueue>* srcMemCopy,
           void* dstBase, void* srcBase) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockSendServer' class.                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockSendServer : public Server {
public :
// wjp change 1998-3-8
  BlockSendServer(List<MessQueue>* sendList, void* _sendBase, int _elementLen) ;

  void transition(MPI_Status* status) ;

  void* active() const ;

private :
  void initSend() ;

  List<MessQueue>* record ;
  int remId ;
  int numMes ;

  int mes ;
  Mess* message ;
  MultiBlock block ;

  void* sendBase ;
  int elementLen ;

//wjp add 1998-3-8
//  PFILE* sendPfile;

  align_t sendBuf [(STATIC_BUFFER_SIZE + ALIGN_SZ - 1) / ALIGN_SZ] ;
} ;

// ----------------------------------------------------------------------
// Things to do
//
//    Lose base addresses from schedules?  Pass them in at schedule
//    execution time.

